(* core.sml *)
(* evaluation and typing *)

structure Core =
struct

open Error
open Syntax

(* ------------------------   EVALUATION  ------------------------ *)

exception NoRuleApplies

fun isnumericval t =
    case t
      of TmZero => true
       | TmSucc t1 => isnumericval t1
       | _ => false

fun isval t =
    case t
      of TmTrue => true
       | TmFalse => true
       | t => isnumericval t

(* eval1:  "call by value" single-step evaluation *)
fun eval1 t =
    case t
      of TmIf(TmTrue,t2,t3) => t2
       | TmIf(TmFalse,t2,t3) => t3
       | TmIf(t1,t2,t3) => TmIf(eval1 t1, t2, t3)
       | TmSucc(t1) => TmSucc(eval1 t1)
       | TmPred(TmZero) => TmZero  (* predicessor of zero is zero *)
       | TmPred(TmSucc(t1)) => 
	   if (isnumericval t1) then t1
	   else TmPred(TmSucc(eval1 t1))
       | TmPred(t1) => TmPred(eval1 t1)
       | TmIsZero(TmZero) => TmTrue
       | TmIsZero(TmSucc(t1)) =>
	   if (isnumericval t1) then TmFalse
           else TmIsZero(TmSucc(eval1 t1))
       | TmIsZero(t1) => TmIsZero(eval1 t1)
       | _ => raise NoRuleApplies

fun eval t =
     let val t' = eval1 t
      in eval t'
     end 
     handle NoRuleApplies => t

(* ------------------------   TYPING  ------------------------ *)

fun typeof t =
    case t
      of TmTrue => TyBool
       | TmFalse => TyBool
       | TmIf(t1,t2,t3) =>
	 if equalTy(typeof t1, TyBool) then
	   let val tyT2 = typeof t2
	    in if equalTy(tyT2, typeof t3) then tyT2
	       else error "arms of conditional have different types"
	   end
	 else error "guard of conditional not a boolean"
       | TmZero => TyNat
       | TmSucc t1 =>
         if equalTy(typeof t1, TyNat) then TyNat
         else error "argument of succ is not a number"
       | TmPred(t1) =>
         if equalTy(typeof t1, TyNat) then TyNat
         else error "argument of pred is not a number"
       | TmIsZero(t1) =>
         if equalTy(typeof t1, TyNat) then TyBool
         else error "argument of iszero is not a number"

end (* structure Core *)
