
structure Syntax =
struct

open PPUtil

(* ---------------------------------------------------------------------- *)
(* Datatypes *)

datatype ty
  = TyBool
  | TyNat

datatype term
  = TmTrue
  | TmFalse
  | TmIf of term * term * term
  | TmZero
  | TmSucc of term
  | TmPred of term
  | TmIsZero of term

(* equality of types *)
fun equalTy (TyBool, TyBool) = true
  | equalTy (TyNat, TyNat) = true
  | equalTy _ = false

(* translating from nonnegative integers to terms *)
fun intToTerm n =
    case n
      of 0 => TmZero
       | n => TmSucc(intToTerm(n-1))


(* ---------------------------------------------------------------------- *)
(* Printing *)

(* The printing functions call these utility functions to insert grouping
  information and line-breaking hints for the pretty-printing library:
     obox   Open a "box" whose contents will be indented by two spaces if
            the whole box cannot fit on the current line
     obox0  Same but indent continuation lines to the same column as the
            beginning of the box rather than 2 more columns to the right
     cbox   Close the current box
     break  Insert a breakpoint indicating where the line maybe broken if
            necessary.
  See the documentation for the Format module in the OCaml library for
  more details. 
*)

open PPUtil

fun obox0() = open_hvbox 0
fun obox() = open_hvbox 2
fun cbox() = close_box()
fun break() = print_break 0 0
val pr = print_string

fun printty_Type outer tyT =
    printty_AType outer tyT

and printty_AType outer tyT =
    case tyT
      of TyBool => pr "Bool"
       | TyNat => pr "Nat"

fun printty tyT = printty_Type true tyT 

fun printtm_Term outer t =
    case t
      of TmIf(t1, t2, t3) =>
	  (obox0();
	   pr "if ";
	   printtm_Term false t1;
	   print_space();
	   pr "then ";
	   printtm_Term false t2;
	   print_space();
	   pr "else ";
	   printtm_Term false t3;
	   cbox())
       | t => printtm_AppTerm outer t

and printtm_AppTerm outer t =
    case t
      of TmPred t1 =>
           (pr "pred "; printtm_ATerm false t1)
       | TmIsZero t1 =>
           (pr "iszero "; printtm_ATerm false t1)
       | t => printtm_ATerm outer t

and printtm_ATerm outer t =
    case t
      of TmTrue => pr "true"
       | TmFalse => pr "false"
       | TmZero => pr "0"
       | TmSucc t1 =>
          let fun f n t =
                  case t
                    of TmZero => pr (Int.toString n)
                     | TmSucc(s) => f (n+1) s
                     | _ => (pr "(succ "; printtm_ATerm false t1; pr ")")
           in f 1 t1
          end
       | t => (pr "("; printtm_Term outer t; pr ")")

fun printtm t = printtm_Term true t 

end (* structure Syntax *)
